import timeit
from typing import Callable
from functools import wraps


def name_from_doc(test: Callable):
    """Get a function's name from it's docstring.

    Either up until the first period or the end of the line, if more
    control is necessary use the register argument.
    """

    d = test.__doc__
    if d is not None:
        return d[:min(d.find("."), d.find("\n"))]
    return None


def timed(name: str = "", printer: Callable[[str], None] = print):
    """Add a timer around a function."""

    def wrapper(func):

        @wraps(func)
        def wrapped(*args, **kwargs):
            start = timeit.default_timer()
            result = func(*args, **kwargs)
            elapsed = timeit.default_timer() - start
            printer(f"{name} finished in {round(elapsed, 5)} seconds")
            return result

        return wrapped
    return wrapper



from typing import Dict, Tuple
from pathlib import Path
from dataclasses import dataclass, field

from ..library import process
from ..library import callgrind

__all__ = ("Resource", "Submission", "Context", "File", "Executable", "ExecutableFile")


class Resource:
    """A resource required for a test."""


@dataclass(eq=False)
class Submission(Resource):
    """The execution context of the tests."""

    problem_path: Path
    assignment_path: Path


@dataclass(eq=False)
class Context(Resource):
    """Extra context."""

    options: Dict[str, str] = field(default_factory=dict)


@dataclass(eq=False)
class File(Resource):
    """A resource corresponding to a file."""

    path: Path


@dataclass(eq=False)
class Executable(Resource):
    """A runnable testing target program."""

    args: Tuple[str, ...]

    def __init__(self, *args: str):
        self.args = args

    def interactive(self, *args: str, cwd: Path = None) -> process.Interactive:
        """Return a subprocess."""

        return process.Interactive(args=self.args + args, cwd=cwd)

    def execute(self, *args: str, stdin: bytes = None, timeout: float = None, cwd: Path = None) -> process.Runtime:
        """Run the target with command line arguments."""

        return process.run(*self.args, *args, stdin=stdin, timeout=timeout, cwd=cwd)

    def count(self, *args: str, stdin: bytes = None, timeout: float = None, cwd: Path = None) -> int:
        """Count the instructions executed during runtime."""

        return callgrind.count(*self.args, *args, stdin=stdin, timeout=timeout, cwd=cwd)


@dataclass(eq=False)
class ExecutableFile(Executable, File):
    """A local file that can be executed."""

    def __init__(self, path: Path, *args: str):
        super().__init__()
        self.path = path
        self.args = (str(path),) + args



import json
import statistics
from typing import List, Dict, Iterable, Union, Set
from dataclasses import dataclass, field
from pathlib import Path

from ...shared import Files


@dataclass
class TaskSummary:
    """Statistics about task results."""

    task: dict
    students_complete: List[dict] = field(default_factory=list)
    students_passing: List[dict] = field(default_factory=list)
    students_timeout: List[dict] = field(default_factory=list)


@dataclass
class StudentProblemSummary:
    """Problem results for a single student."""

    tasks_complete: List[dict] = field(default_factory=list)
    tasks_passing: List[dict] = field(default_factory=list)


@dataclass
class StudentSummary:
    """Statistics for an individual student."""

    student: dict
    problems: Dict[str, StudentProblemSummary]

    def __init__(self, student: dict, problem_shorts: Iterable[str]):
        self.student = student
        self.problems = {}
        for problem_short in problem_shorts:
            self.problems[problem_short] = StudentProblemSummary()


@dataclass
class ProblemSummary:
    """Statistics about a set of tests."""

    tasks: Dict[str, TaskSummary]

    def __init__(self, tasks: dict):
        self.tasks = {task_short: TaskSummary(task) for task_short, task in tasks.items()}


@dataclass
class Summary:
    """Overall assignment statistics."""

    problems: Dict[str, ProblemSummary]
    students: Dict[str, StudentSummary]
    failed_setup: Set[str]

    def __init__(self, grading_schema: dict, students: dict):
        """Generate skeleton of data store."""

        self.problems = {}
        for problem_short, data in grading_schema["problems"].items():
            self.problems[problem_short] = ProblemSummary(data["tasks"])
        self.students = {}
        problem_shorts = tuple(grading_schema["problems"].keys())
        for student_username, student in students.items():
            self.students[student_username] = StudentSummary(student, problem_shorts)
        self.failed_setup = set()


def build_student_summary(summary: Summary, report_path: Path):
    """Build a table of results with axes students and tasks."""

    report_name = report_path.parts[-1].rsplit(".")[0]
    student_summary = summary.students[report_name]

    with report_path.open() as file:
        report = json.load(file)

    for problem_short, problem_report in report.items():
        problem_summary = summary.problems[problem_short]

        for task_name, result in problem_report.items():
            task_summary = problem_summary.tasks[task_name]
            task = task_summary.task
            if task["stage"] == "setup" and not result["passing"]:
                summary.failed_setup.add(report_name)

            if result["complete"]:
                task_summary.students_complete.append(student_summary.student)
                student_summary.problems[problem_short].tasks_complete.append(task)
            if result["passing"]:
                task_summary.students_passing.append(student_summary.student)
                student_summary.problems[problem_short].tasks_passing.append(task)
            if "runtime" in result and result["runtime"] is not None and result["runtime"]["timeout"] is not None:
                task_summary.students_timeout.append(student_summary.student)

    return student_summary


def build_summary(grading_schema: dict, report_paths: Iterable[Path]) -> Summary:
    """Compile problem summaries."""

    students = {}
    for report_path in report_paths:
        student_username = report_path.parts[-1].rsplit(".")[0]
        students[student_username] = dict(username=student_username)
    summary = Summary(grading_schema, students)
    for report_path in report_paths:
        build_student_summary(summary, report_path)
    return summary


def percent(x: Union[int, float], n: int = 1) -> str:
    return f"{round(x / n * 1000) / 10 if n != 0 else 0}%"


def filter_tests(tasks: List[dict]) -> List[dict]:
    return list(task for task in tasks if task["stage"] == "test")


def summarize(grading_path: Path, report_paths: Iterable[Path]):
    """Do summaries of the reports in a directory."""

    with grading_path.joinpath(Files.GRADING).open() as file:
        grading_schema = json.load(file)

    summary = build_summary(grading_schema, report_paths)

    for problem_short, problem_summary in summary.problems.items():
        print(f"Problem: {problem_short}")

        print("  Tests")
        for task_name, task_summary in problem_summary.tasks.items():
            print(f"    {task_name}: {len(task_summary.students_passing)}/{len(task_summary.students_complete)}",
                  f"({len(task_summary.students_timeout)} timeout)")

        scores = []
        for student_username, student_summary in summary.students.items():
            count_tests_complete = len(filter_tests(student_summary.problems[problem_short].tasks_complete))
            if count_tests_complete:
                count_tests_passing = len(filter_tests(student_summary.problems[problem_short].tasks_passing))
                scores.append(count_tests_passing / count_tests_complete)
                # print(student_username, 100 * count_tests_passing / count_tests_complete)

        print("  Statistics")
        print(f"    Total scores: {len(scores)}")
        print(f"    Mean: {percent(statistics.mean(scores)) if len(scores) > 0 else '-'}")
        print(f"    Median: {percent(statistics.median(scores)) if len(scores) > 0 else '-'}")
        print(f"    Perfect: {percent(len(list(filter(lambda x: x == 1, scores))), len(scores))}")
        # print(f"    Scores: {list(scores)}")

    print(f"Submissions that failed setup: {len(summary.failed_setup)}")
    for report_name in summary.failed_setup:
        print(f"  {report_name}")



from typing import Optional
from dataclasses import dataclass, field
from pathlib import Path

from ...task import Result


@dataclass
class CodeLocation:
    """File, line, column."""

    path: Path
    line: Optional[int] = None
    column: Optional[int] = None

    def __str__(self):
        if self.column:
            return f"{self.path.parts[-1]}:{self.line}.{self.column}"
        if self.line:
            return f"{self.path.parts[-1]}:{self.line}"
        return f"{self.path}"


@dataclass(init=False)
class CodeResult(Result):
    """The result of a correctness case."""

    description: str = field(init=False)
    location: CodeLocation = field(init=False)

    def __init__(self, complete: bool, passing: bool, description: str, location: CodeLocation, **details):
        super().__init__(complete=complete, passing=passing, details=details)
        self.description = description
        self.location = location



from .grader import Grader
from .task import Error
from .setup.build import BuildResult
from .setup.check import CheckResult
from .test.correctness import CorrectnessResult
from .test.complexity import ComplexityResult
from .test.memory import MemoryResult
from .teardown.cleanup import CleanupResult
from .resource import Submission, Context, File, Executable, ExecutableFile
from .setup import SetupResult
from .teardown import TeardownResult



"""Grade is a combined library for grading C++ problem sets.

Grade is capable of testing three main aspects of C++ programs:
correctness, efficiency, and style. It includes tools for testing the
output of the code under various conditions, measuring its runtime
performance and memory usage, and running style checks.
"""

import timeit

from pathlib import Path
from typing import Iterator, Iterable, Tuple

from ..log import log
from .models import GradingAssignment, GradingProblem
from .report import AssignmentReport
from .resource import Submission, Context


root = Path(__file__).absolute().parent
include = root.joinpath("include")


def run(assignment: GradingAssignment, submission_path: Path, options: dict) -> AssignmentReport:
    """Run all tests on a submission and return a dict of results."""

    log.info(f"running {submission_path}")
    reports = AssignmentReport.create(assignment)

    start = timeit.default_timer()

    context = Context(options=options)
    for problem in filter(lambda p: p.grading.is_automated, assignment.problems):
        log.debug(f"running problem {problem.short}")
        submission = Submission(
            assignment_path=submission_path,
            problem_path=submission_path.joinpath(problem.relative_path))
        reports[problem.short] = problem.grader.run(context=context, submission=submission)

    elapsed = timeit.default_timer() - start
    log.debug(f"finished in {round(elapsed, 5)} seconds")

    return reports


def run_batch(self, target_paths: Iterable[Path], options: dict) -> Iterator[Tuple[Path, AssignmentReport]]:
    """Run multiple reports, map directory to report."""

    # Start timer
    start = timeit.default_timer()

    for target_path in target_paths:
        yield target_path, self.run_single(target_path, options=options)

    elapsed = timeit.default_timer() - start
    log.info(f"finished batch in {round(elapsed, 5)} seconds")



from ..task import Result
from ..stage import GraderStage, Registrar
from .cleanup import CleanupResult


class TeardownResult(Result):
    """Common result for generic tasks."""

    kind = "generic"

    def __init__(self, passing: bool = True, complete: bool = True, **details):
        super().__init__(complete=complete, passing=passing, details=details)


class TeardownStage(GraderStage):
    """Teardown endpoints."""

    name = "teardown"

    def generic(self, **details) -> Registrar:
        """Generic tasks."""

        return self.registrar(details, TeardownResult)

    def cleanup(self, **details) -> Registrar:
        """Deleting files."""

        return self.registrar(details, CleanupResult)



import math

from dataclasses import dataclass
from typing import Sequence, Callable, Iterable


@dataclass()
class RuntimeDataPoint:
    n: float
    count: float


class RuntimeDataTransform:
    """Transforms for 2D data."""

    @classmethod
    def n_log_n(cls, point: RuntimeDataPoint) -> RuntimeDataPoint:
        return RuntimeDataPoint(math.log(point.n), point.count)


@dataclass(eq=False)
class RuntimeData:
    points: Sequence[RuntimeDataPoint]

    @classmethod
    def build(cls, raw: Iterable[Iterable[float]]) -> "RuntimeData":
        return RuntimeData(tuple(RuntimeDataPoint(*point) for point in sorted(raw, key=lambda p: p[0])))

    def transform(self, profile: Callable[[RuntimeDataPoint], RuntimeDataPoint]) -> "RuntimeData":
        return RuntimeData(tuple(map(profile, self.points)))

    def linearity(self) -> float:
        """Calculate linear regression error.

        This method computes the linear regression of the data,
        uses the slope and intercept to determine error, then returns
        a factor describing how close to linear the data set is.

        This was developed by Jamie Smith. From his notes:

        > In my preliminary testing, I found that comparing the
        > average error and average y^2 values was a reasonable test
        > for how linear the data is. If the average y^2 is an order
        > of magnitude or two greater than the average error, the data
        > is more or less linear. This ratio also has the nice
        > property that if you multiply all times by a scalar, it
        > stays the same, so if the user's function gets 10x faster in
        > every case, the correlation factor doesn't change at all.
        """

        sum_xy = sum_x = sum_y = sum_x_squared = sum_y_squared = 0
        for point in self.points:
            sum_xy += point.n * point.count
            sum_x += point.n
            sum_y += point.count
            sum_x_squared += point.n * point.n
            sum_y_squared += point.count * point.count

        m = (len(self.points) * sum_xy - sum_x * sum_y) / (len(self.points) * sum_x_squared - sum_x * sum_x)
        b = (sum_y - m * sum_x) / len(self.points)

        sum_error_squared = 0
        for point in self.points:
            expected = m * point.n + b
            sum_error_squared += (expected - point.count) ** 2

        average_error_squared = sum_error_squared / len(self.points)
        average_y_squared = sum_y_squared / len(self.points)
        if average_error_squared == 0:
            return float("inf")
        return math.log10(average_y_squared) - math.log10(average_error_squared)



from typing import List, Tuple, Iterator, Set, Optional, Callable
from dataclasses import dataclass, field

from .report import ProblemReport
from .task import Task, Result
from .stage import GraderStage
from .exception import GraderException
from ..log import log

from .resource import Context, Submission
from .dependency import topological_sort
from .setup import SetupStage
from .test import TestStage
from .teardown import TeardownStage

import typing

if typing.TYPE_CHECKING:
    from .models import GradingProblem


def fulfills_dependencies(task: Task, report: ProblemReport):
    """Convenience."""

    return all((
        all(report[dependency].passing for dependency in task.dependencies.passing),
        all(report[dependency].complete for dependency in task.dependencies.complete)))


@dataclass(eq=False, init=False)
class TaskFilter:
    """Small helper to check whether a task should be run."""

    tags: Optional[Set[str]] = None
    task_names: Optional[Set[str]] = None

    def __init__(self, context: Context, problem_short: str):
        """Build from context."""

        tags = context.options.get("tags")
        if tags is not None:
            self.tags = set(tags)
            TaskFilter.filter_problem_specific(self.tags, problem_short)
        task_names = context.options.get("tasks")
        if task_names is not None:
            self.task_names = set(task_names)
            TaskFilter.filter_problem_specific(self.task_names, problem_short)

    def __call__(self, task: Task) -> bool:
        """Check if a task should be run."""

        if self.tags is not None:
            if self.tags.isdisjoint(task.tags):
                return False
        if self.task_names is not None:
            if task.name not in self.task_names:
                return False
        return True

    @property
    def has_effect(self) -> bool:
        return self.tags is not None or self.task_names is not None

    @staticmethod
    def filter_problem_specific(collection: Set[str], prefix: str):
        """Filter in items prefaced by prefix:xyz as xyz."""

        for item in filter(lambda i: ":" in i, collection):
            collection.remove(item)
            if item.startswith(f"{prefix}:"):
                collection.add(item.split(":", maxsplit=1)[1])


def _run(
        tasks: List[Task],
        is_visible: Callable[[Task], bool],
        resources: dict,
        report: ProblemReport):
    """Execute sorted tasks, skipping if missing dependencies."""

    log.debug("running tasks")
    for task in tasks:
        log.debug(f"running task {task.name}")

        # Check conditions for whether this case is filtered out, if so report is partial
        if not is_visible(task):
            report.partial = True
            continue

        # If we can't run it, mark as incomplete
        elif not fulfills_dependencies(task, report):
            result = task.Result.incomplete()

        # Run task if not hidden and dependencies are met
        else:
            try:
                result = task.run(resources)

            # Results may be raised
            except Result as r:
                result = r

            # Check if the result is the right type
            if not isinstance(result, task.Result):
                raise GraderException(f"expected result type {task.Result.kind} from {task.name} in {task.source}")

        result.task = task
        report.add(result)


@dataclass(eq=False)
class Grader:
    """A main class for grading runtime."""

    setup: SetupStage = field(default_factory=SetupStage)
    test: TestStage = field(default_factory=TestStage)
    teardown: TeardownStage = field(default_factory=TeardownStage)

    # Populated on import
    problem: "GradingProblem" = field(init=False)

    @property
    def stages(self) -> Tuple[GraderStage, ...]:
        return self.setup, self.test, self.teardown

    @property
    def tasks(self) -> Iterator[Task]:
        yield from self.setup.tasks
        yield from self.test.tasks
        yield from self.teardown.tasks

    def check(self):
        """Topologically sort tasks, checking for cycles."""

        # Check dependencies
        log.debug("sorting grader tasks by dependency")
        topological_sort(stage.tasks for stage in self.stages)

    def run(self, context: Context, submission: Submission) -> ProblemReport:
        """Build and test."""

        log.debug("setting up runtime")

        # Resources
        resources = dict(context=context, submission=submission)
        resources.update(resources=resources)

        # Create the filter
        is_visible = TaskFilter(context, self.problem.short)

        # Final report
        report = ProblemReport.create(self.problem)

        # Run each stage
        for stage in self.stages:
            if len(stage.tasks) > 0:
                log.debug(f"starting stage {stage.name}")
                _run(stage.tasks, is_visible, resources, report)
            else:
                log.debug(f"no tasks for stage {stage.name}")

        return report



from library import valgrind
from . import MemoryResult


def make_test_memory(executable_name: str, *args: str, timeout: float = 1):
    """Make the actual memory test."""

    def test_memory(resources: dict):
        executable = resources[executable_name]
        return MemoryResult.from_valgrind_report(valgrind.run(*executable.args, *args, timeout=timeout))

    return test_memory



import textwrap

class Printer:
    """Helper class for conveniently printing."""

    def __init__(self):
        self.buffer = []
        self.indentation = 0

    def print(self, *args, sep: str = " ", end: str = "\n", indentation: int = 0):
        """Standard print API."""

        self.buffer.append(textwrap.indent(sep.join(args), " " * (indentation + self.indentation)) + end)

    def indent(self, amount: int = 2):
        self.indentation += amount

    def dedent(self, amount: int = 2):
        self.indentation = max(0, self.indentation - amount)

    def __str__(self):
        return "".join(self.buffer)



class Singleton(type):
    """Singleton metaclass."""

    __instances = {}

    def __call__(cls, *args, **kwargs):
        """Instantiate if not in the map, otherwise return existing."""

        if cls not in cls.__instances:
            cls.__instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.__instances[cls]



from pathlib import Path
from typing import Iterator

__all__ = ("version", "Paths", "Files", "Templates")


version = "0.1.0"


class Paths:
    DOT = Path()
    ASSIGNMENT = Path("assignment")
    PROBLEM = Path("problem")
    ARTIFACTS = Path("artifacts")
    INSTRUCTIONS = Path("instructions")
    RESOURCES = Path("resources")
    SOLUTION = Path("solution")
    GRADING = Path("grading")
    ASSETS = Path("assets")
    INCLUDE = Path("grade", "include")

    @classmethod
    def glob_assignments(cls, material_path: Path) -> Iterator[Path]:
        """Provide a unified search for assignments."""

        for path in material_path.joinpath(cls.ASSIGNMENT).glob("*/"):
            if path.is_dir():
                yield path


class Files:
    README = "README.md"
    ASSIGNMENT = "assignment.json"
    PROBLEM = "problem.json"
    GRADING = "grading.json"
    TESTS = "tests.py"
    INDEX = "index.json"


class Templates:
    ASSIGNMENT = "assignment.md"
    PROBLEM = "problem.md"



from typing import Optional
from pathlib import Path

from ...task import Error
from . import CheckResult

__all__ = ("check_file_exists", "check_makefile_exists", "search_file_by_name")


def check_file_exists(*paths: Path) -> CheckResult:
    """Check if a file is present in the directory."""

    if not any(path.exists() for path in paths):
        return CheckResult(passing=False, error=Error(description=f"can't find {paths[0].parts[-1]}"))
    return CheckResult(passing=True)


def check_makefile_exists(path: Path) -> CheckResult:
    """Check whether there is a makefile in a directory."""

    lower_path = path.joinpath("makefile")
    upper_path = path.joinpath("Makefile")

    if not lower_path.exists() and not upper_path.exists():
        return CheckResult(passing=False, error=Error(description=f"can't find {upper_path.parts[-1]}"))
    return CheckResult(passing=True)


def search_file_by_name(name: str, path: Path) -> Optional[Path]:
    """Find file by name, None if none or multiple."""

    results = tuple(path.glob(f"**/{name}"))
    if len(results) == 1:
        return results[0]
    return None



import json

from pathlib import Path
from typing import List, Optional
from dataclasses import field

from ..models import Assignment, Problem
from ..shared import Files
from .exception import CompilationException


class CompilationProblem(Problem):
    """Add additional fields only used for build."""

    number: Optional[int]
    path: Path
    assignment: "CompilationAssignment"

    percentage: float = None

    @property
    def index_path(self) -> Path:
        return self.path.joinpath(Files.PROBLEM)

    @classmethod
    def read(
            cls,
            assignment: "CompilationAssignment",
            reference: dict,
            root: Path,
            number: int = None) -> "CompilationProblem":
        """Load a problem from the assignment path and reference."""

        path = root.joinpath(reference["path"])
        index_path = path.joinpath(Files.PROBLEM)

        try:
            with index_path.open() as file:
                data = json.load(file)
        except FileNotFoundError:
            raise CompilationException(message=f"No such path {index_path} for assignment {assignment.path}")
        except (PermissionError, OSError):
            raise CompilationException(message=f"Failed to read {index_path} for assignment {assignment.path}")
        except json.decoder.JSONDecodeError:
            raise CompilationException(message=f"Failed to deserialize {index_path} for assignment {assignment.path}")

        data["short"] = reference.get("short", data.get("short", path.parts[-1]))
        data["relative_path"] = reference.get("relative_path", data["short"])

        if "title" in reference:
            data["title"] = reference["title"]

        data["grading"]["enabled"] = reference["grading"].get("enabled", True)
        data["grading"]["weight"] = reference["grading"].get("weight", "1")
        data["grading"]["points"] = reference["grading"].get("points", "100")
        for category in "automated", "review", "manual":
            category_data = data["grading"][category]
            if category_data is None:
                continue

            if category in reference["grading"]:
                reference_category_data = reference["grading"][category]

                if "enabled" in reference_category_data:
                    category_data["enabled"] = reference_category_data
                if "weight" in reference_category_data:
                    category_data["weight"] = reference_category_data["weight"]
                if "points" in reference_category_data:
                    category_data["points"] = reference_category_data["points"]

            if "weight" not in category_data:
                category_data["weight"] = "1"
            if "points" not in category_data:
                category_data["points"] = "100"

        self = cls.load(data)

        # Convenience details for rendering
        self.assignment = assignment
        self.number = number
        self.path = path

        return self


class CompilationAssignment(Assignment):
    """Additional fields for build."""

    problems: List[CompilationProblem]
    path: Path = field(init=False)

    @property
    def index_path(self) -> Path:
        return self.path.joinpath(Files.ASSIGNMENT)

    @classmethod
    def read(cls, path: Path) -> "CompilationAssignment":
        """Load an assignment from a containing directory."""

        index_path = path.joinpath(Files.ASSIGNMENT)

        try:
            with index_path.open() as file:
                data = json.load(file)
        except FileNotFoundError:
            raise CompilationException(message=f"No such path {index_path}")
        except (PermissionError, OSError):
            raise CompilationException(message=f"Failed to read {index_path}")
        except json.decoder.JSONDecodeError:
            raise CompilationException(message=f"Failed to deserialize {index_path}")

        data["short"] = data.get("short", path.parts[-1])
        self = cls.load(data, problems=[])

        counter = 1
        total_weight = 0
        for reference in data.pop("problems"):
            problem = CompilationProblem.read(self, reference, path)
            if problem.grading.is_automated or problem.grading.is_review or problem.grading.is_manual:
                problem.number = counter
                counter += 1

            total_weight += problem.grading.weight
            self.problems.append(problem)

        for problem in self.problems:
            problem.percentage = problem.grading.weight / total_weight if total_weight > 0 else 0

        self.path = path

        return self



from __future__ import annotations

import typing
import datetime
from typing import Dict, Optional
from dataclasses import dataclass, field

from .task import Result
from ..models import serialize_datetime, deserialize_datetime

if typing.TYPE_CHECKING:
    from .models import GradingAssignment, GradingProblem


@dataclass(eq=False)
class ProblemReportProblemReference:
    """Reference to original problem."""

    short: str
    title: str

    def dump(self) -> dict:
        return dict(short=self.short, title=self.title)

    @classmethod
    def create(cls, problem: "GradingProblem") -> "ProblemReportProblemReference":
        return ProblemReportProblemReference(short=problem.short, title=problem.title)

    @classmethod
    def load(cls, data: dict) -> "ProblemReportProblemReference":
        return ProblemReportProblemReference(**data)


@dataclass(eq=False)
class ProblemReport:
    """The final report returned by the testing framework."""

    problem: ProblemReportProblemReference
    results: Dict[str, Result] = field(default_factory=dict)
    partial: bool = False

    def __getitem__(self, item: str) -> Result:
        """Look up a result by task name."""

        return self.results[item]

    def get(self, item: str) -> Optional[Result]:
        """Mimic lookup get."""

        return self.results.get(item)

    def add(self, result: Result):
        """Add a result to the report."""

        self.results[result.task.name] = result

    def dump(self, thin: bool = False) -> dict:
        """Dump the result to a serializable format."""

        results = {result.task.name: result.dump(thin=thin) for result in self.results.values()}
        return dict(problem=self.problem.dump(), partial=self.partial, results=results)

    @classmethod
    def create(cls, problem: GradingProblem) -> "ProblemReport":
        """Create a new problem."""

        return ProblemReport(problem=ProblemReportProblemReference.create(problem))

    @classmethod
    def load(cls, data: dict, problem: GradingProblem) -> "ProblemReport":
        """Deserialize, rebinding to provided tasks."""

        partial = data["partial"]
        results = {}
        for task in problem.grader.tasks:
            result_data = data["results"].get(task.name)
            if result_data is not None:
                results[task.name] = Result.load(result_data, task)
            else:
                partial = True

        return ProblemReport(
            problem=ProblemReportProblemReference.load(data["problem"]),
            results=results,
            partial=partial)


@dataclass(eq=False)
class AssignmentReportAssignmentReference:
    """Structured data about the origin assignment."""

    short: str
    title: str
    # hash: str

    def dump(self) -> dict:
        return dict(short=self.short, title=self.title)

    @classmethod
    def create(cls, assignment: GradingAssignment) -> "AssignmentReportAssignmentReference":
        return AssignmentReportAssignmentReference(short=assignment.short, title=assignment.title)

    @classmethod
    def load(cls, data: dict):
        return AssignmentReportAssignmentReference(**data)


@dataclass(eq=False)
class AssignmentReport:
    """Aggregation of problem reports."""

    assignment: AssignmentReportAssignmentReference
    problems: Dict[str, ProblemReport] = field(default_factory=dict)
    timestamp: datetime.datetime = field(default_factory=datetime.datetime.now)
    partial: bool = False

    def __post_init__(self):
        """Check if any initialized problems are partial."""

        for problem in self.problems.values():
            if problem.partial:
                self.partial = True
                break

    def __contains__(self, item: str) -> bool:
        return item in self.problems

    def __getitem__(self, item: str) -> ProblemReport:
        """Index problem reports by problem short."""

        return self.problems[item]

    def __setitem__(self, key: str, value: ProblemReport):
        """Set the result from a problem."""

        self.problems[key] = value
        self.partial = self.partial or value.partial

    def dump(self, thin: bool = False) -> dict:
        """Serialize as dictionary to shorten rebuild."""

        return dict(
            assignment=self.assignment.dump(),
            timestamp=serialize_datetime(self.timestamp),
            partial=self.partial,
            problems={short: report.dump(thin=thin) for short, report in self.problems.items()})

    @classmethod
    def create(cls, assigment: "GradingAssignment") -> "AssignmentReport":
        """Create from assignment for metadata."""

        return AssignmentReport(assignment=AssignmentReportAssignmentReference.create(assigment))

    @classmethod
    def load(cls, data: dict, assignment: "GradingAssignment") -> "AssignmentReport":
        """Deserialize and bind to existing tasks."""

        assignment_reference = AssignmentReportAssignmentReference.load(data.pop("assignment"))
        timestamp = deserialize_datetime(data.pop("timestamp"))

        problems = {}
        for problem in assignment.problems:
            if problem.grading.is_automated:
                problems[problem.short] = ProblemReport.load(data["problems"][problem.short], problem)

        return AssignmentReport(
            assignment=assignment_reference,
            problems=problems,
            timestamp=timestamp,
            partial=data["partial"])



import linecache
import tracemalloc


def summarize(snapshot: tracemalloc.Snapshot, key_type: str, limit: int):
    """Summarize snapshot in console."""

    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_statistics = snapshot.statistics(key_type)

    for i, statistic in enumerate(top_statistics[:limit], 1):
        frame = statistic.traceback[0]
        print("#%s: %s:%s: %.1f KiB" % (i, frame.filename, frame.lineno, statistic.size / 1024))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print("    %s" % line)

    other = top_statistics[limit:]
    if other:
        size = sum(statistic.size for statistic in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(statistic.size for statistic in top_statistics)
    print("Total allocated size: %.1f KiB" % (total / 1024))



import importlib.util
import json
import sys
from pathlib import Path
from typing import List
from dataclasses import field
from decimal import Decimal
from functools import lru_cache

from ..models import Assignment, Problem, ProblemGrading, Author
from ..shared import Files
from .grader import Grader


def import_grader(grading_path: Path, problem: "GradingProblem", grader_name: str = "grader") -> Grader:
    """Import a grader from a tests file."""

    # Try to import as a module
    if grading_path.joinpath("__init__.py").is_file():
        name = f"_curricula_grading_{grading_path.parts[-1]}"
        spec = importlib.util.spec_from_file_location(name, str(grading_path.joinpath("__init__.py")))
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module

    # Otherwise there's a tests.py
    else:
        name = f"_curricula_grading_{grading_path.parts[-1]}_tests"
        spec = importlib.util.spec_from_file_location(name, str(grading_path.joinpath("tests.py")))
        module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)
    grader = getattr(module, grader_name)
    grader.problem = problem

    return grader


class GradingProblemGrading(ProblemGrading):
    """Override to add grader-specific computations."""

    problem: "GradingProblem"

    @property
    @lru_cache(maxsize=1)
    def point_ratio(self) -> Decimal:
        return self.points / self.problem.grader.test.weight


class GradingProblem(Problem):
    """Additional details for grading."""

    path: Path = field(init=False)
    grader: Grader = field(init=False)
    grading: GradingProblemGrading

    @classmethod
    def read(cls, data: dict, path: Path) -> "GradingProblem":
        """Import the grader."""

        self = GradingProblem.load(data)
        self.path = path
        if self.grading.is_automated:
            self.grader = import_grader(path, self)
        return self


class GradingAssignment(Assignment):
    """Additional details for grading."""

    path: Path

    problems: List[GradingProblem]

    @classmethod
    def read(cls, path: Path) -> "GradingAssignment":
        """Read the assignment.json and rebuild the models."""

        with path.joinpath(Files.INDEX).open("r") as file:
            data = json.load(file)

        problems = []
        for problem_data in data.pop("problems"):
            problems.append(GradingProblem.read(
                data=problem_data,
                path=path.joinpath(problem_data["short"])))

        assignment = GradingAssignment.load(data, problems=problems)
        assignment.path = path

        return assignment



import json
from typing import Any, TextIO


def truncate(string: str, length: int, append: str = "...") -> str:
    """Shorthand for cutting off long strings.

    If length is zero or negative, the string will not be checked or
    truncated. The original string will be returned.
    """

    if len(string) > length > 0:
        return string[:length - len(append)] + append
    return string


def descend_and_truncate(o: Any, length: int, append: str = "..."):
    """Truncate strings in a JSON object."""

    if isinstance(o, str):
        return truncate(o, length, append)
    if isinstance(o, dict):
        for key, value in o.items():
            o[key] = descend_and_truncate(value, length, append)
    if isinstance(o, list):
        for i, item in enumerate(o):
            o[i] = descend_and_truncate(item, length, append)
    return o


def dump(o: Any, file: TextIO, no_truncate: bool = False, **options):
    """Write an object to a file."""

    if not no_truncate:
        descend_and_truncate(o, 100_000)
    json.dump(o, file, **options)


def load(file: TextIO):
    """Read data from a file."""

    return json.load(file)



from ...task import Result, Error


class ComplexityResult(Result):
    """The result of a correctness case."""

    kind = "correctness"

    def __init__(self, passing: bool, complete: bool = True, error: Error = None, **details):
        super().__init__(complete=complete, passing=passing, error=error, details=details)



import stat
from typing import Iterable, Optional, Tuple
from pathlib import Path

from . import BuildResult
from ...resource import ExecutableFile, File
from ...task import Error
from ....library import process
from ....library.files import delete_file, add_mode

__all__ = ("build_gpp_executable", "build_makefile_executable", "build_harness_library")


def build_gpp_executable(
        source_path: Path,
        *source_paths: Path,
        destination_path: Path,
        gpp_options: Iterable[str] = (),
        timeout: float = None) -> Tuple[BuildResult, Optional[ExecutableFile]]:
    """Build a binary from a single C++ file with G++."""

    destination_path.parent.mkdir(parents=True, exist_ok=True)
    runtime = process.run(
        "g++", *gpp_options,
        "-o", str(destination_path),
        str(source_path),
        *map(str, source_paths),
        timeout=timeout)

    error_description = None
    error_traceback = None
    if runtime.raised_exception:
        error_description = f"error invoking compilation of {source_path.parts[-1]}: {runtime.exception.description}"
    elif runtime.timed_out:
        error_description = f"timed out while compiling {source_path.parts[-1]}"
    elif runtime.code != 0:
        if runtime.stderr:
            error_description = "failed to compile"
            error_traceback = runtime.stderr.decode(errors="replace")
        else:
            error_description = "nonzero status code during compilation"
    elif not destination_path.exists():
        error_description = f"build did not produce {destination_path.parts[-1]}"

    # If the build failed
    if error_description is not None:
        error = Error(description=error_description, traceback=error_traceback)
        return BuildResult(passing=False, runtime=runtime.dump(), error=error), None

    # Chmod
    add_mode(destination_path, stat.S_IXOTH)

    # Otherwise
    return BuildResult(passing=True, runtime=runtime.dump()), ExecutableFile(destination_path)


def build_makefile_executable(
        target_path: Path,
        make_options: Iterable[str] = (),
        timeout: float = None) -> BuildResult:
    """Run make on the target directory."""

    runtime = process.run("make", "-B", "-C", str(target_path), *make_options, timeout=timeout)
    if runtime.code != 0 or runtime.timed_out:
        error = f"failed to make {target_path.parts[-1]}"
        return BuildResult(passing=False, runtime=runtime.dump(), error=Error(description=error))
    return BuildResult(passing=True, runtime=runtime.dump())


def build_harness_library(
        harness_path: Path,
        include_paths: Iterable[Path] = (),
        object_paths: Iterable[Path] = (),
        gpp_options: Iterable[str] = (),
        build_path: Path = Path(),
        timeout: int = 30) -> Tuple[BuildResult, Optional[File]]:
    """Build a shared object file.

    The target path will be listed after an include flag so that
    headers may be imported.
    """

    object_path = build_path.joinpath("harness.o")
    runtime = process.run(
        "g++", "-Wall", "-c", "-fPIC", *gpp_options,
        str(harness_path.absolute()),
        *map(lambda path: f"{path.absolute()}", object_paths),
        *map(lambda path: f"-I{path.absolute()}", include_paths),
        "-o", str(object_path),
        timeout=timeout)
    if runtime.code != 0 or runtime.raised_exception is not None:
        return BuildResult(passing=False, runtime=runtime.dump(), error=Error(description="compilation failed")), None

    shared_object_path = build_path.joinpath("harness.so")
    runtime = process.run("g++", "-shared", str(object_path), "-o", str(shared_object_path), timeout=timeout)
    delete_file(object_path)
    if runtime.code != 0 or runtime.raised_exception is not None:
        return (
            BuildResult(
                passing=False,
                runtime=runtime.dump(),
                error=Error(description="shared library build failed")),
            None)

    return BuildResult(passing=True), File(shared_object_path)



import abc
from decimal import Decimal

from ..task import Result
from ..stage import GraderStage
from .code import CodeResult
from .complexity import ComplexityResult
from .correctness import CorrectnessResult
from .memory import MemoryResult


class Test(abc.ABC):
    """Convenience class for building dynamic test objects."""

    def __call__(self, *args, **kwargs) -> Result:
        """Should behave like a standard runnable."""


class TestStage(GraderStage):
    """Test endpoints."""

    name = "test"

    @property
    def weight(self) -> Decimal:
        """Get the cumulative weight of all tasks."""

        return sum(task.weight for task in self.tasks)

    def code(self, **details):
        """Code quality."""

        return self.registrar(details, CodeResult)

    def complexity(self, **details):
        """Runtime complexity."""

        return self.registrar(details, ComplexityResult)

    def correctness(self, **details):
        """Program correctness."""

        return self.registrar(details, CorrectnessResult)

    def memory(self, **details):
        """Resource allocation."""

        return self.registrar(details, MemoryResult)



from pathlib import Path

from ..models import CompilationAssignment
from ..framework import Workflow, TargetResult
from ...shared import Paths, Files
from ...log import log
from ...library.files import copy_directory, move


class SiteWorkflow(Workflow):
    """Grades the solution and writes the report to the artifact root."""

    def run(self, assignment: CompilationAssignment, result: TargetResult):
        """Main."""

        if not self.configuration.options.get("site"):
            return

        if not result["instructions"].compiled:
            return

        site_path = Path(self.configuration.options["site"])
        if not site_path.parent.exists():
            log.error("site parent path does not exist, cancelling pipeline")
            return

        log.info(f"pipelining instructions to site {site_path}")
        copy_directory(
            source=self.configuration.artifacts_path.joinpath(Paths.INSTRUCTIONS),
            destination=site_path,
            merge=False)
        move(site_path.joinpath(Files.README), site_path.joinpath("index.md"))



from pathlib import Path

from ..library.utility import timed
from ..log import log

from .compilation import CurriculaTarget, Configuration


@timed("compile", printer=log.info)
def compile(assignment_path: Path, artifacts_path: Path, custom_template_path: Path = None, options: dict = None):
    """Build the assignment at a given path."""

    configuration = Configuration(
        assignment_path=assignment_path,
        artifacts_path=artifacts_path,
        custom_template_path=custom_template_path,
        options=options)
    target = CurriculaTarget(configuration)
    target.compile()



from dataclasses import dataclass

from ...task import Error, Result


@dataclass(init=False, eq=False)
class BuildResult(Result):
    """Returned from a build task."""

    kind = "build"

    def __init__(self, passing: bool, complete: bool = True, error: Error = None, **details):
        super().__init__(complete=complete, passing=passing, error=error, details=details)



from typing import List, Type, Callable
from decimal import Decimal

from .task import Task, Runnable, Dependencies, Result
from .exception import GraderException
from ..library.debug import get_source_location

__all__ = ("GraderStage", "Registrar")

Registrar = Callable[[Runnable], Runnable]
RegistrarFactory = Callable[[dict, Type[Result]], Registrar]


class GraderStage:
    """Management for setup, test, and teardown stages."""

    name: str
    tasks: List[Task]

    def __init__(self):
        """Initialize new grader stage with empty tasks."""

        self.tasks = []

    def registrar(self, details: dict, result_type: Type[Result]):
        """Generate a registrar for tasks.

        This is to be used internally within named registration
        endpoints and returns a decorator.
        """

        def decorator(runnable: Runnable) -> Runnable:
            """Put the function in a correctness object."""

            name = details.pop("name", None) or runnable.__qualname__
            description = details.pop("description", None) or runnable.__doc__
            weight = Decimal(details.pop("weight", 1))
            dependencies = Dependencies.from_details(details)
            tags = details.pop("tags", set())

            for existing_task in self.tasks:
                if existing_task.name == name:
                    raise GraderException(f"Duplicate task name \"{name}\"")

            # Create task, append
            self.tasks.append(Task(
                name=name,
                description=description,
                stage=self.name,
                kind=result_type.kind,
                dependencies=dependencies,
                runnable=runnable,
                details=details,
                weight=weight,
                source=get_source_location(2),
                tags=tags,
                Result=result_type))
            return runnable

        return decorator



from .task import Task
from .exception import GraderException

from typing import Dict, List, Iterable

__all__ = ("topological_sort",)


def topological_sort_visit(task: Task, lookup: Dict[str, Task], marks: Dict[Task, int], result: List[Task]):
    """Visit a node."""

    if marks[task] == 2:
        return

    if marks[task] == 1:
        raise GraderException("found cycle in task dependencies")

    marks[task] = 1
    for dependency in task.dependencies.passing | task.dependencies.complete:
        topological_sort_visit(lookup[dependency], lookup, marks, result)
    marks[task] = 2
    result.append(task)


def topological_sort(stage_tasks: Iterable[List[Task]]):
    """Order tasks by dependency."""

    lookup = {}
    marks = {}
    for tasks in stage_tasks:
        result = []
        for task in tasks:
            marks[task] = 0
            lookup[task.name] = task
        for task in tasks:
            if marks[task] != 2:
                topological_sort_visit(task, lookup, marks, result)
        tasks.clear()
        tasks.extend(result)





import jinja2

from dataclasses import dataclass
from pathlib import Path
from typing import List, Callable, Union, Optional, Tuple, TypeVar

from ..shared import Files, Templates
from ..library import files
from ..log import log

from .models import CompilationProblem, CompilationAssignment
from .framework import Configuration, Context

__all__ = (
    "get_readme_path",
    "get_readme",
    "has_readme",
    "ReadmeBuilder",
    "DirectoryMerger",
    "DirectoryAggregator")


def get_readme_path(relative_path: Union[str, Path] = None):
    """Return the expected README.md path."""

    if relative_path is None:
        return Path(Files.README)
    if isinstance(relative_path, str):
        relative_path = Path(relative_path)
    return relative_path.joinpath(Files.README)


@jinja2.environmentfilter
def get_readme(
        environment: jinja2.Environment,
        item: Union[CompilationProblem, CompilationAssignment],
        relative_path: str = None) -> str:
    """Render a README with options for nested path."""

    readme_path = get_readme_path(relative_path).as_posix()

    try:
        if isinstance(item, CompilationAssignment):
            path = f"assignment:{readme_path}"
            data = dict(assignment=item)
        elif isinstance(item, CompilationProblem):
            path = f"problem/{item.short}:{readme_path}"
            data = dict(assignment=item.assignment, problem=item)
        else:
            raise ValueError("invalid item passed to get_readme")
        return environment.get_template(path).render(**data)

    except jinja2.exceptions.TemplateNotFound:
        log.error(f"error finding problem/{item.short}:{readme_path}")
        return ""


def has_readme(item: Union[CompilationProblem, CompilationAssignment], relative_path: str = None) -> bool:
    """Check whether a problem has a solution README."""

    return item.path.joinpath(relative_path or "", Files.README).exists()


T = TypeVar("T")


@dataclass(eq=False)
class ReadmeBuilder:
    """Compose a README based on an assignment and its problems."""

    configuration: Configuration
    readme_relative_path: Path
    template_relative_path: Path
    destination_path: Path

    def __post_init__(self):
        """Set destination to exact path."""

        if self.destination_path.parts[-1].lower() != Files.README.lower():
            self.destination_path = self.destination_path.joinpath(Files.README)

    def run(self, assignment: CompilationAssignment, context: Context) -> Path:
        """Render the template returning the written path."""

        log.debug(f"building {self.readme_relative_path}/README.md to {self.destination_path}")
        template_path = self.template_relative_path.joinpath(Templates.ASSIGNMENT)
        template = context.environment.get_template(f"template:compile/{template_path.as_posix()}")
        with self.destination_path.open("w") as file:
            file.write(template.render(assignment=assignment))
        return self.destination_path

    def should_run(self, assignment: CompilationAssignment, context: Context):
        """Check if any paths are used by this builder."""

        if context.paths_modified is None or context.indices_modified:
            return True

        # Check if a template has been modified
        for path in context.paths_modified:
            if files.contains(self.configuration.custom_template_path, path):
                return True

        # Check if a README has been modified
        readme_path = get_readme_path(self.readme_relative_path)
        used_paths = {assignment.path.joinpath(readme_path).resolve()}
        for problem in assignment.problems:
            used_paths.add(problem.path.joinpath(readme_path).resolve())

        return len(context.paths_modified.intersection(used_paths)) > 0

    def run_if_should(self, assignment: CompilationAssignment, context: Context) -> Tuple[bool, Optional[Path]]:
        """Run if any test_paths are used."""

        if self.should_run(assignment, context):
            return True, self.run(assignment, context)
        return False, None


class DirectoryShouldRun:
    """Shared uses for merger and aggregator."""

    contents_relative_path: Path
    run: Callable[[CompilationAssignment], T]
    should_run: Callable[[CompilationAssignment, Context], bool]

    def should_run(self, assignment: CompilationAssignment, context: Context):
        """Check if any paths are used by this builder."""

        if context.paths_modified is None or context.indices_modified:
            return True

        paths = [assignment.path.joinpath(self.contents_relative_path)]
        for problem in assignment.problems:
            paths.append(assignment.path.joinpath(problem.relative_path, self.contents_relative_path).absolute())

        for path_modified in context.paths_modified:
            for path in paths:
                if files.contains(path, path_modified):
                    return True

        return False

    def run_if_should(self, assignment: CompilationAssignment, context: Context) -> Tuple[bool, T]:
        """Run if any test_paths are used."""

        if self.should_run(assignment, context):
            return True, self.run(assignment)
        return False, None


@dataclass(eq=False)
class DirectoryMerger(DirectoryShouldRun):
    """Merges a set of directories into one."""

    contents_relative_path: Path
    destination_path: Path
    filter_problems: Optional[Callable[[CompilationProblem], bool]] = None

    def run(self, assignment: CompilationAssignment):
        """Merge directory contents."""

        log.debug(f"merging directories {self.contents_relative_path} to {self.destination_path}")

        files.replace_directory(self.destination_path)

        # First copy any assignment-wide resources
        assignment_contents_path = assignment.path.joinpath(self.contents_relative_path)
        if assignment_contents_path.is_dir():
            files.copy_directory(assignment_contents_path, self.destination_path)

        # Overwrite with problem contents, enable filtration
        for problem in filter(self.filter_problems, assignment.problems):
            problem_contents_path = problem.path.joinpath(self.contents_relative_path)
            if problem_contents_path.is_dir():
                files.copy_directory(problem_contents_path, self.destination_path, merge=True)


@dataclass(eq=False)
class DirectoryAggregator(DirectoryShouldRun):
    """Aggregates directories into a single parent."""

    contents_relative_path: Path
    destination_path: Path
    filter_problems: Optional[Callable[[CompilationProblem], bool]] = None
    directory_name: Optional[Callable[[CompilationProblem], str]] = None

    def run(self, assignment: CompilationAssignment) -> List[Path]:
        """Aggregate contents."""

        log.debug(f"aggregating {self.contents_relative_path} to {self.destination_path}")

        files.replace_directory(self.destination_path)

        # First compile assignment-wide assets
        assignment_contents_path = assignment.path.joinpath(self.contents_relative_path)
        if assignment_contents_path.exists():
            files.copy_directory(assignment_contents_path, self.destination_path)

        directory_name = self.directory_name or (lambda p: p.relative_path)

        # Copy per problem, enable filtration
        copied_paths = []
        for problem in filter(self.filter_problems, assignment.problems):
            problem_contents_path = problem.path.joinpath(self.contents_relative_path)
            if problem_contents_path.exists():
                problem_destination_path = self.destination_path.joinpath(directory_name(problem))
                copied_paths.append(problem_destination_path)
                files.copy_directory(problem_contents_path, problem_destination_path)

        return copied_paths



from ..stage import GraderStage, Registrar
from ..task import Result, Error
from .check import CheckResult
from .build import BuildResult


class SetupResult(Result):
    """Common result for generic tasks."""

    kind = "generic"

    def __init__(self, passing: bool = True, complete: bool = True, error: Error = None, **details):
        super().__init__(complete=complete, passing=passing, error=error, details=details)


class SetupStage(GraderStage):
    """Setup endpoints."""

    name = "setup"

    def generic(self, **details) -> Registrar:
        """Generic tasks."""

        return self.registrar(details, SetupResult)

    def build(self, **details) -> Registrar:
        """Compilation."""

        return self.registrar(details, BuildResult)

    def check(self, **details) -> Registrar:
        """Passive checks."""

        return self.registrar(details, CheckResult)



import jinja2

import json
from typing import Dict, Iterable, Iterator, List
from pathlib import Path
from dataclasses import dataclass, field
from decimal import Decimal

from ..task import Result
from ..models import GradingAssignment, GradingProblem
from ..report import AssignmentReport, ProblemReport
from ...library.template import jinja2_create_environment, DEFAULT_TEMPLATE_PATH, pretty


def sum_weights(results: Iterable[Result]) -> Decimal:
    return sum(result.task.weight for result in results)


@dataclass(eq=False)
class ProblemSummary:
    """A summary of the results from a problem's cases."""

    problem: GradingProblem
    report: ProblemReport

    # Setup issues
    setup_results_errored: List[Result] = field(default_factory=list)

    # Main tests
    test_results_count: int = 0
    test_results_passing_count: int = 0

    # Weight
    test_results_passing_weight: Decimal = Decimal(0)

    def __post_init__(self):
        """Cache some common analysis of the data."""

        for task in self.problem.grader.tasks:
            result = self.report[task.name]

            if task.stage == "setup":
                if result.error is not None:
                    self.setup_results_errored.append(result)

            if task.stage == "test":
                self.test_results_count += 1

                # Increment counts
                if result.passing and result.complete:
                    self.test_results_passing_count += 1
                    self.test_results_passing_weight += task.weight

    @property
    def test_results_passing(self) -> Iterator[Result]:
        """Count the number of tests that passed."""

        for task in self.problem.grader.tasks:
            if task.stage == "test":
                result = self.report[task.name]
                if result.passing and result.complete:
                    yield result

    @property
    def test_results_failing(self) -> Iterator[Result]:
        """Count the number of tests that passed."""

        for task in self.problem.grader.tasks:
            if task.stage == "test":
                result = self.report[task.name]
                if not result.passing or not result.complete:
                    yield result

    @property
    def passing_fraction(self) -> str:
        return f"{self.test_results_passing_count}/{self.test_results_count}"

    @property
    def points_fraction(self) -> str:
        """Format a fraction."""

        numerator = self.test_results_passing_weight
        denominator = self.problem.grader.test.weight
        if denominator == 0:
            return f"0/0"

        points_total_automated = self.problem.grading.automated.points
        if denominator != points_total_automated:
            numerator = numerator / denominator * points_total_automated
            denominator = points_total_automated
        return f"{pretty(numerator)}/{pretty(denominator)}"

    @property
    def points_percentage(self) -> Decimal:
        """Compute the percentage."""

        if self.test_results_count == 0:
            return Decimal("0")

        numerator = self.test_results_passing_weight
        denominator = self.problem.grader.test.weight
        if denominator == 0:
            return Decimal("0")

        return Decimal(numerator) / Decimal(denominator)


@dataclass(eq=False)
class ReportSummary:
    """Summary of problems."""

    problems: Dict[str, ProblemSummary] = field(default_factory=dict)


def summarize(assignment: GradingAssignment, report: AssignmentReport) -> ReportSummary:
    """Do a summary for a single student."""

    summary = ReportSummary()
    for problem in assignment.problems:
        if problem.grading.is_automated:
            summary.problems[problem.short] = ProblemSummary(problem, report[problem.short])

    return summary


def create_format_environment(custom_template_path: Path = None) -> jinja2.Environment:
    """Create the requisite environment."""

    if custom_template_path is None:
        custom_template_path = DEFAULT_TEMPLATE_PATH
    return jinja2_create_environment(custom_template_path=custom_template_path)


def format_report_markdown(
        assignment: GradingAssignment,
        report_path: Path,
        environment: jinja2.Environment = None,
        options: dict = None) -> str:
    """Return a formatted markdown report."""

    with report_path.open() as file:
        report = AssignmentReport.load(json.load(file), assignment)
    if report.partial:
        return "Cannot format a partial report!"

    summary = summarize(assignment, report)
    if environment is None:
        environment = create_format_environment()

    environment.globals.update(assignment=assignment, summary=summary, options=options)
    report_template = environment.get_template("template:grade/report/assignment.md")
    return report_template.render() + "\n"



from typing import Optional

from ....library.process import Runtime
from ....library.valgrind import ValgrindReport
from ...task import Result, Error


class MemoryResult(Result):
    """Result of a memory leak test."""

    kind = "memory"

    runtime: Optional[Runtime]
    error_count: Optional[int]
    leaked_blocks: Optional[int]
    leaked_bytes: Optional[int]

    def __init__(
            self,
            passing: bool,
            runtime: Runtime = None,
            complete: bool = True,
            error_count: int = None,
            leaked_blocks: int = None,
            leaked_bytes: int = None,
            **details):
        """Initialize a memory result."""

        super().__init__(complete=complete, passing=passing, details=details)
        self.runtime = runtime
        self.error_count = error_count
        self.leaked_blocks = leaked_blocks
        self.leaked_bytes = leaked_bytes

        # Create the error with additional data
        if error_count > 0 or leaked_bytes > 0:
            self.error = Error(description=f"leaked {leaked_bytes} bytes with {error_count} errors")

    @classmethod
    def from_valgrind_report(cls, report: ValgrindReport, **details) -> "MemoryResult":
        """Generate a memory test result from a valgrind report."""

        if report.errors is None:
            return cls(complete=False, passed=False, runtime=report.runtime)
        leaked_blocks, leaked_bytes = report.memory_lost()
        return cls(
            passed=leaked_bytes == 0 and len(report.errors) == 0,
            runtime=report.runtime,
            error_count=len(report.errors),
            leaked_blocks=leaked_blocks,
            leaked_bytes=leaked_bytes,
            **details)

    def dump(self) -> dict:
        dump = super().dump()
        dump.update(
            runtime=self.runtime.dump() if self.runtime else None,
            error_count=self.error_count,
            leaked_blocks=self.leaked_blocks,
            leaked_bytes=self.leaked_bytes)
        return dump



from typing import Callable, Iterable
from pathlib import Path

from ..task import Error
from . import SetupResult
from ...library.process import TimeoutExpired


def make_open_interactive(
        executable_name: str,
        count: int,
        args: Iterable[str] = (),
        cwd: Path = None,
        read_condition: Callable[[bytes], bool] = None,
        read_condition_timeout: float = None):
    """Make a method that opens interactive executables."""

    def test(resources: dict):
        try:
            for i in range(count):
                name = f"{executable_name}_i{i}"
                resources[name] = interactive = resources[executable_name].interactive(*args, cwd=cwd)
                if read_condition is not None:
                    interactive.stdout.read(condition=read_condition, timeout=read_condition_timeout)
        except OSError:
            return SetupResult(passing=False, error=Error(description="failed to open process"))
        except TimeoutExpired:
            return SetupResult(passing=False, error=Error(description=f"{executable_name} timed out"))

    return test



import abc

from typing import *
from pathlib import Path

from . import CorrectnessResult
from .. import Test
from ...task import Error
from ....library.process import Runtime, Interactive, TimeoutExpired, Interaction
from ....library.configurable import none, not_none, Configurable

__all__ = (
    "as_lines",
    "lines_match",
    "lines_match_unordered",
    "test_runtime_succeeded",
    "OutputTest",
    "CompareTest",
    "CompareBytesOutputTest",
    "ExecutableMixin",
    "ExecutableCodeMixin",
    "ExecutableInputFileMixin",
    "ExecutableOutputFileMixin",
    "CompareExitCodeOutputTest",
    "write_then_read")


def as_lines(string: AnyStr) -> List[AnyStr]:
    """Strip and split by newline."""

    return string.strip().split("\n" if isinstance(string, str) else b"\n")


AnyStrSequence = Union[Iterable[AnyStr], Sized]


def lines_match(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check ordered equality.

    Returns a boolean indicating correctness and a list of errors
    encountered while checking.
    """

    if hasattr(a, "__len__") and hasattr(a, "__len__"):
        if len(a) != len(b):
            return False

    for x, y in zip(a, b):
        if x != y:
            return False

    return True


def lines_match_unordered(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check unordered equality."""

    return lines_match(sorted(a), sorted(b))


CompareTest = Callable[[Any], CorrectnessResult]


class OutputTest(Test, abc.ABC):
    """Compares program output to an expected output."""

    test: CompareTest
    resources: Optional[dict]
    details: Optional[dict]

    def __init__(self, **kwargs):
        """Create shared fields."""

        if len(kwargs) > 0:
            raise SyntaxError(f"base class received kwargs {kwargs.keys()}")

    def get_output(self) -> Any:
        """Override this."""

        pass

    def __call__(self, resources: dict) -> CorrectnessResult:
        """Check if the output matches."""

        self.resources = resources
        self.details = {}
        result = self.test(self.get_output())
        result.details.update(self.details)
        self.resources = None
        self.details = None
        return result


BytesTransform = Callable[[bytes], bytes]
T = TypeVar("T")


def identity(x: T) -> T:
    return x


class CompareBytesOutputTest(OutputTest, Configurable):
    """Compares output to expected values."""

    out_transform: BytesTransform
    out_line_transform: BytesTransform
    test_out: bytes
    test_out_lines: Iterable[bytes]
    test_out_lines_lists: Iterable[Iterable[bytes]]

    def __init__(
            self,
            *,
            out_transform: BytesTransform = none,
            out_line_transform: BytesTransform = none,
            test_out: bytes = none,
            test_out_lines: Iterable[bytes] = none,
            test_out_lines_lists: Iterable[Iterable[bytes]] = none,
            **kwargs):
        """Build a test for matching stdout output.

        The list of lines from the stdout of the runtime is compared
        against each list of lines in test_out_line_lists. Note that this
        method first checks whether any error was raised during runtime.
        """

        super().__init__(**kwargs)
        self.test_out = self.resolve("test_out", local=test_out, default=None)
        self.test_out_lines = self.resolve("test_out_lines", local=test_out_lines, default=None)
        self.test_out_lines_lists = self.resolve("test_out_lines_lists", local=test_out_lines_lists, default=None)

        # Check resolvable
        resolvable = tuple(filter(None, (self.test_out, self.test_out_lines, self.test_out_lines_lists)))
        if len(resolvable) != 1:
            raise ValueError("Exactly one of test_out, test_out_lines, or test_out_lines_lists is required")

        self.out_transform = self.resolve("out_transform", local=out_transform, default=identity)
        self.out_line_transform = self.resolve("out_line_transform", local=out_line_transform, default=identity)

    def test(self, out: bytes):
        """Call the corresponding test."""

        if self.test_out is not None:
            return self._test_out(out)
        return self._test_out_lines_lists(out)

    def _test_out(self, out: bytes) -> CorrectnessResult:
        """Shortcut compare for one option block text."""

        test_out = self.resolve("test_out")

        out = self.out_transform(out)
        passing = out == test_out
        error = None if passing else Error(
            description="unexpected output",
            expected=test_out.decode(errors="replace"),
            received=out.decode(errors="replace"))
        return CorrectnessResult(passing=passing, error=error)

    def _test_out_lines_lists(self, out: bytes) -> CorrectnessResult:
        """Used to compare multiple options of lists of lines."""

        if self.test_out_lines is not None:
            test_out_lines_lists = [self.test_out_lines]
        else:
            test_out_lines_lists = self.test_out_lines_lists

        out_lines = tuple(map(self.out_line_transform, self.out_transform(out).split(b"\n")))
        passing = any(lines_match(out_lines, lines) for lines in test_out_lines_lists)
        expected = tuple(b"\n".join(test_out_lines).decode(errors="replace") for test_out_lines in test_out_lines_lists)
        error = None if passing else Error(
            description="unexpected output",
            expected=expected[0] if len(expected) == 1 else expected,
            received=b"\n".join(out_lines).decode(errors="replace"))
        return CorrectnessResult(passing=passing, error=error)


def test_runtime_succeeded(runtime: Runtime) -> CorrectnessResult:
    """See if the runtime raised exceptions or returned status code."""

    if runtime.raised_exception:
        error = Error(description=runtime.exception.description)
        raise CorrectnessResult(passing=False, runtime=runtime.dump(), error=error)
    if runtime.timed_out:
        error = Error(
            description="timed out",
            suggestion=f"expected maximum elapsed time of {runtime.timeout} seconds")
        raise CorrectnessResult(passing=False, runtime=runtime.dump(), error=error)
    if runtime.code != 0:
        error = Error(
            description=f"received status code {runtime.code}",
            suggestion="expected status code of zero")
        return CorrectnessResult(passing=False, runtime=runtime.dump(), error=error)


class ExecutableMixin(OutputTest, Configurable):
    """Meant for reading and comparing stdout."""

    executable_name: str
    args: Iterable[str]
    stdin: Optional[bytes]
    timeout: Optional[float]
    cwd: Optional[Path]

    def __init__(
            self,
            *,
            executable_name: str = none,
            args: Iterable[str] = none,
            stdin: bytes = none,
            timeout: float = none,
            cwd: Path = none,
            **kwargs):
        """Save container information, call super."""

        super().__init__(**kwargs)
        self.executable_name = executable_name
        self.args = args
        self.stdin = stdin
        self.timeout = timeout
        self.cwd = cwd

    def execute(self) -> Runtime:
        """Check that it ran correctly, then run the test."""

        executable = self.resources[self.executable_name]
        runtime = executable.execute(
            *self.resolve("args", default=()),
            stdin=self.resolve("stdin", default=None),
            timeout=self.resolve("timeout", default=None),
            cwd=self.resolve("cwd", default=None))
        self.details["runtime"] = runtime.dump()
        return runtime

    def get_output(self) -> Any:
        """Return stdout by default."""

        runtime = self.execute()
        test_runtime_succeeded(runtime)
        return runtime.stdout


class ExecutableCodeMixin(ExecutableMixin):
    """Output is exit code instead of stdout."""

    def get_output(self) -> Any:
        """Return exit code."""

        return self.execute().code


class ExecutableInputFileMixin(OutputTest, Configurable):
    """Enables the use of an input file rather than stdin string."""

    stdin: bytes
    input_file_path: Path

    def __init__(self, *, input_file_path: Path = none, **kwargs):
        """Set stdin."""

        super().__init__(**kwargs)
        self.input_file_path = input_file_path
        self.stdin = self.resolve("input_file_path").read_bytes()


class ExecutableOutputFileMixin(OutputTest, Configurable):
    """Enables the use of an input file rather than stdin string."""

    output_file_path: Path

    def __init__(self, *, output_file_path: Path = none, **kwargs):
        """Set stdin."""

        super().__init__(**kwargs)
        self.output_file_path = output_file_path

    def get_output(self) -> Any:
        """Call super because it might do something."""

        super().get_output()
        output_file_path = self.resolve("output_file_path")
        if not output_file_path.is_file():
            raise CorrectnessResult(passing=False, error=Error(
                description="no output file",
                suggestion=f"expected path {output_file_path}"))
        return output_file_path.read_bytes()


CompareNativeTest = Callable[[Any], CorrectnessResult]


class CompareExitCodeOutputTest(OutputTest, Configurable):
    """Checks program exit code."""

    test: CompareNativeTest
    resources: Optional[dict]
    details: Optional[dict]

    expected_code: int
    expected_codes: Container[int]

    def __init__(self, *, expected_code: int = none, expected_codes: Container[int] = none, **kwargs):
        """Set expected codes."""

        super().__init__(**kwargs)

        expected_code = self.resolve("expected_code", local=expected_code, default=None)
        expected_codes = self.resolve("expected_codes", local=expected_codes, default=None)

        if expected_code is None and expected_codes is None:
            raise ValueError("Runtime exit test requires either expected status or statuses!")
        if expected_code is not None:
            expected_codes = (expected_code,)

        def test(code: int):
            if code in expected_codes:
                return CorrectnessResult(passing=True)
            return CorrectnessResult(
                passing=False,
                error=Error(
                    description=f"received incorrect exit code f{code}",
                    suggestion=f"expected {expected_codes}"))

        self.test = test


def write_then_read(
        interactive: Interactive,
        stdin: Sequence[bytes],
        read_condition: Callable[[bytes], bool] = None,
        read_condition_timeout: float = None) -> Interaction:
    """Pass in a bunch of lines, get the last output, compare."""

    if not interactive.poll():
        raise CorrectnessResult(passing=False, error=Error(description="program crashed"))

    with interactive.recording() as interaction:
        for command in stdin:
            interactive.stdin.write(command)
            try:
                interactive.stdout.read(
                    block=True,
                    condition=read_condition,
                    timeout=read_condition_timeout)
            except TimeoutExpired:
                raise CorrectnessResult(
                    passing=False,
                    error=Error(description="timed out"),
                    runtime=interaction.dump())

    if interaction.stdout is None:
        raise CorrectnessResult(
            passing=False,
            error=Error(description="did not receive output"),
            runtime=interaction.dump())

    return interaction



"""A package that provides tools for building assignments.

Curricula's build system takes an assignment composed of a set of a
set of problems, and separates the contents of each into combined
instructions, resources, grading, and solution artifacts. The
structure of an assignment is as follows:

assignment/
| README.md
| assignment.json

The structure of a problem is somewhat similar:

problem/
| README.md
| problem.json
| assets/
| | ...
| grading/
| | README.md
| | tests.py
| | ...
| resources/
| | ...
| solution/
| | README.md
| | ...

Typically, problems are located in a sub-folder assignment/problem/.
However, the problems are linked to assignments in the assignment.json
file via relative path, so the problem directory could be anywhere.
"""

import json
from pathlib import Path

from ..shared import Files, Paths
from ..library.template import jinja2_create_environment
from ..library import files
from ..log import log

from .validate import validate
from .content import ReadmeBuilder, DirectoryMerger, DirectoryAggregator, get_readme, has_readme
from .models import CompilationAssignment
from .framework import Unit, UnitResult, Context, Configuration, Target, TargetResult
from .workflows.grade import GradeWorkflow
from .workflows.site import SiteWorkflow

root = Path(__file__).parent.absolute()


class InstructionsUnit(Unit):
    """Builds the instructions and assets."""

    output_path: Path
    readme_builder: ReadmeBuilder
    asset_merger: DirectoryMerger

    name = "instructions"
    README = "readme"
    ASSETS = "assets"

    def __init__(self, configuration: Configuration):
        """Create builders."""

        super().__init__(configuration)
        self.output_path = self.configuration.artifacts_path.joinpath(Paths.INSTRUCTIONS)

        self.readme_builder = ReadmeBuilder(
            configuration=self.configuration,
            readme_relative_path=Paths.DOT,
            template_relative_path=Paths.INSTRUCTIONS,
            destination_path=self.output_path)
        self.asset_merger = DirectoryMerger(
            contents_relative_path=Paths.ASSETS,
            destination_path=self.output_path.joinpath(Paths.ASSETS))

    def compile_readme(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Combine instructions."""

        run, _ = self.readme_builder.run_if_should(assignment, context)
        log.info(f"""instructions: {"built" if run else "skipped"} readme""")
        if run:
            result.compiled = True
            result.tags.add(self.README)

    def merge_assets(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Combine instruction assets into one folder."""

        run, _ = self.asset_merger.run_if_should(assignment, context)
        log.info(f"""instructions: {"merged" if run else "skipped"} assets""")
        if run:
            result.compiled = True
            result.tags.add(self.ASSETS)

    def compile(self, assignment: CompilationAssignment, context: Context) -> UnitResult:
        """Create directory and add README and assets."""

        result = UnitResult()
        self.output_path.mkdir(exist_ok=True)
        self.compile_readme(assignment, context, result)
        self.merge_assets(assignment, context, result)
        return result


class ResourcesUnit(Unit):
    """Builds the instructions and assets."""

    output_path: Path
    content_aggregator: DirectoryAggregator

    name = "resources"
    RESOURCES = "resources"

    def __init__(self, configuration: Configuration):
        """Only a content aggregator here."""

        super().__init__(configuration)
        self.output_path = self.configuration.artifacts_path.joinpath(Paths.RESOURCES)
        self.content_aggregator = DirectoryAggregator(
            contents_relative_path=Paths.RESOURCES,
            destination_path=self.output_path,
            directory_name=lambda p: p.relative_path)

    def compile(self, assignment: CompilationAssignment, context: Context) -> UnitResult:
        """Bring the problem resources into one directory."""

        result = UnitResult()
        run, _ = self.content_aggregator.run_if_should(assignment, context)
        log.info(f"""resources: {"aggregated" if run else "skipped"} resources""")
        if run:
            result.compiled = True
            result.tags.add(self.RESOURCES)
        return result


class SolutionUnit(Unit):
    """Creates cheat sheet and solution code."""

    output_path: Path
    readme_builder: ReadmeBuilder
    content_aggregator: DirectoryAggregator

    name = "solution"
    README = "readme"
    CONTENTS = "contents"

    def __init__(self, configuration: Configuration):
        """Configure and setup builders."""

        super().__init__(configuration)
        self.output_path = self.configuration.artifacts_path.joinpath(Paths.SOLUTION)

        self.readme_builder = ReadmeBuilder(
            configuration=self.configuration,
            readme_relative_path=Paths.SOLUTION,
            template_relative_path=Paths.SOLUTION,
            destination_path=self.output_path)
        self.content_aggregator = DirectoryAggregator(
            contents_relative_path=Paths.SOLUTION,
            destination_path=self.output_path,
            directory_name=lambda p: p.relative_path)

    def compile_readme(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Assemble cheat sheet."""

        run, _ = self.readme_builder.run_if_should(assignment, context)
        log.info(f"""solution: {"built" if run else "skipped"} readme""")
        if run:
            result.compiled = True
            result.tags.add(self.README)

    def aggregate_contents(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Match resources with solutions."""

        run, copied_paths = self.content_aggregator.run_if_should(assignment, context)
        log.info(f"""solution: {"merged" if run else "skipped"} contents""")

        if run:
            # Delete extra READMEs
            for copied_path in copied_paths:
                readme_path = copied_path.joinpath(Files.README)
                if readme_path.exists():
                    files.delete(readme_path)

            result.compiled = True
            result.tags.add(self.CONTENTS)

    def compile(self, assignment: CompilationAssignment, context: Context) -> UnitResult:
        """Run and get result."""

        result = UnitResult()
        self.output_path.mkdir(exist_ok=True)
        self.compile_readme(assignment, context, result)
        self.aggregate_contents(assignment, context, result)
        return result


class GradingUnit(Unit):
    """Assemble grading scripts and rubric."""

    output_path: Path
    readme_builder: ReadmeBuilder
    content_aggregator: DirectoryAggregator

    name = "grading"
    README = "readme"
    CONTENTS = "contents"
    INDEX = "index"

    def __init__(self, configuration: Configuration):
        """Configure and setup builders."""

        super().__init__(configuration)
        self.output_path = self.configuration.artifacts_path.joinpath(Paths.GRADING)

        self.readme_builder = ReadmeBuilder(
            configuration=self.configuration,
            readme_relative_path=Paths.GRADING,
            template_relative_path=Paths.GRADING,
            destination_path=self.output_path)
        self.content_aggregator = DirectoryAggregator(
            contents_relative_path=Paths.GRADING,
            destination_path=self.output_path,
            filter_problems=lambda p: p.grading.automated,
            directory_name=lambda p: p.short)

    def compile_readme(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Assemble rubric."""

        run, _ = self.readme_builder.run_if_should(assignment, context)
        log.info(f"""solution: {"built" if run else "skipped"} readme""")

        if run:
            result.compiled = True
            result.tags.add(self.README)

    def aggregate_contents(self, assignment: CompilationAssignment, context: Context, result: UnitResult):
        """Get all grading scripts."""

        run, copied_paths = self.content_aggregator.run_if_should(assignment, context)
        log.info(f"""solution: {"merged" if run else "skipped"} contents""")

        if run:
            # Delete extra READMEs
            for copied_path in copied_paths:
                readme_path = copied_path.joinpath(Files.README)
                if readme_path.exists():
                    files.delete(readme_path)

            result.compiled = True
            result.tags.add(self.CONTENTS)

    def dump_index(self, assignment: CompilationAssignment, result: UnitResult):
        """Dump the assignment for grading reference."""

        with self.output_path.joinpath(Files.INDEX).open("w") as file:
            json.dump(assignment.dump(), file, indent=2)
        result.compiled = True
        result.tags.add(self.INDEX)

    def compile(self, assignment: CompilationAssignment, context: Context):
        """Compile rubric, grading scripts, and index."""

        result = UnitResult()
        self.output_path.mkdir(exist_ok=True)
        self.compile_readme(assignment, context, result)
        self.aggregate_contents(assignment, context, result)
        self.dump_index(assignment, result)
        return result


class CurriculaTarget(Target):
    """Repeatable build instructions."""

    configuration: Configuration

    def __init__(self, configuration: Configuration):
        """Initialize the compilation and do static setup."""

        super().__init__(configuration)

        # Register compilation units
        self.unit(InstructionsUnit)
        self.unit(ResourcesUnit)
        self.unit(SolutionUnit)
        self.unit(GradingUnit)

        # Register additional workflows
        self.workflow(GradeWorkflow)
        self.workflow(SiteWorkflow)

    def compile(self, **context_options) -> TargetResult:
        """Generate context and compile."""

        log.info(f"building {self.configuration.assignment_path} to {self.configuration.artifacts_path}")

        # Validate first
        validate(self.configuration.assignment_path)

        # Load the assignment object
        log.debug("loading assignment")
        assignment = CompilationAssignment.read(self.configuration.assignment_path)

        # Set up templating
        problem_template_paths = {f"problem/{problem.short}": problem.path for problem in assignment.problems}
        environment = jinja2_create_environment(
            assignment_path=self.configuration.assignment_path,
            problem_paths=problem_template_paths,
            custom_template_path=self.configuration.custom_template_path)
        environment.filters.update(get_readme=get_readme, has_readme=has_readme)
        environment.globals["configuration"] = self.configuration

        # Define context
        log.debug("setting context")
        context = Context(environment=environment, **context_options)

        # Check if an index file was edited
        if context.paths_modified:
            for item in (assignment, *assignment.problems):
                if item.index_path.resolve() in context.paths_modified:
                    context.indices_modified = True
                    break

        # Create output directory
        self.configuration.artifacts_path.mkdir(exist_ok=True, parents=True)

        # Run units
        return super().compile(assignment, context)



import abc
import inspect
from typing import Dict, TypeVar, Generic, Any, Type, Set
from dataclasses import dataclass, field, asdict
from decimal import Decimal

from ..log import log

__all__ = ("Error", "Result", "Dependencies", "Task", "Runnable")


@dataclass(eq=False)
class Error:
    """An error raised during a task."""

    description: str
    suggestion: str = None
    location: str = None
    traceback: str = None
    expected: Any = None
    received: Any = None

    def dump(self, thin: bool = False) -> dict:
        if thin:
            return dict(description=self.description, suggestion=self.suggestion)
        return asdict(self)

    @classmethod
    def load(cls, data: dict) -> "Error":
        return cls(**data)


@dataclass(init=False, eq=False)
class Result(Exception, abc.ABC):
    """The result of a test."""

    complete: bool
    passing: bool
    details: dict
    error: Error

    kind: str = field(init=False)
    task: "Task" = field(init=False, repr=False)

    def __init__(self, complete: bool, passing: bool, error: Error = None, details: dict = None):
        """Initialize a new result.

        Details are passed as a dictionary in order to avoid potential
        collisions with normal arguments.
        """

        self.complete = complete
        self.passing = passing
        self.error = error
        self.details = details or dict()

    def dump(self, thin: bool = False) -> dict:
        """Serialize the result for JSON."""

        dump = dict(
            complete=self.complete,
            passing=self.passing,
            kind=self.kind,
            error=self.error.dump(thin=thin) if self.error is not None else self.error,
            task=dict(
                name=self.task.name,
                description=self.task.description))
        if not thin:
            dump.update(details=self.details)
        return dump

    @classmethod
    def load(cls, data: dict, task: "Task") -> "Result":
        """Load a result from serialized."""

        data.pop("task")
        kind = data.pop("kind")
        error_data = data.pop("error")
        error = Error.load(error_data) if error_data is not None else None
        self = cls(**data, error=error)
        self.task = task
        self.kind = kind
        return self

    @classmethod
    def incomplete(cls):
        """Return a mock result if the task was not completed."""

        return cls(complete=False, passing=False)

    @classmethod
    def default(cls):
        """Called in special cases if no result is returned."""

        return cls(complete=True, passing=True)


TResult = TypeVar("TResult", bound=Result)


class Runnable(Generic[TResult]):
    """Runnable function that returns some kind of result."""

    def __call__(self, **kwargs) -> TResult:
        """The signature that will receive dependency injection."""


@dataclass(eq=False)
class Dependencies:
    """Task dependencies based on passing or completion."""

    passing: Set[str]
    complete: Set[str]

    @classmethod
    def normalize_from_details(cls, name: str, details: dict) -> Set[str]:
        """Normalize a set of strings."""

        value = details.pop(name, None)
        if value is None:
            return set()
        if isinstance(value, str):
            return {value}
        if isinstance(value, set):
            return value
        return set(value)

    @classmethod
    def from_details(cls, details: dict):
        """Parse from decorator details."""

        return cls(
            passing=cls.normalize_from_details("passing", details),
            complete=cls.normalize_from_details("complete", details))

    def dump(self):
        return dict(passing=list(self.passing), complete=list(self.complete))


@dataclass(eq=False)
class Task(Generic[TResult]):
    """Superclass for check, build, run."""

    name: str
    description: str
    stage: str
    kind: str

    dependencies: Dependencies
    runnable: Runnable[Result]
    details: dict

    weight: Decimal
    source: str
    tags: Set[str]

    Result: Type[Result]

    def run(self, resources: Dict[str, Any]) -> TResult:
        """Do the dependency injection for the runnable."""

        dependencies = {}
        for name, parameter in inspect.signature(self.runnable).parameters.items():
            dependency = resources.get(name, parameter.default)
            if dependency == parameter.empty:
                raise ValueError(f"caught in {self.name}: could not satisfy dependency {name}")
            dependencies[name] = dependency
        result = self.runnable(**dependencies)

        if result is None:
            log.debug(f"task {self.name} did not return a result")
            return self.Result.default()

        return result



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="curricula",
    version="1.2.4",
    description="A content manager and grading toolkit for evaluating student code",
    url="https://github.com/csci104/curricula",
    author="Noah Kim",
    author_email="noahbkim@gmail.com",

    # Extra
    long_description=long_description,
    long_description_content_type="text/markdown",

    # Python
    python_requires=">=3.7",

    # Packaging
    packages=find_packages(),
    package_data={"curricula": [
        "compile/schema/*.json",
        "grade/include/*.hpp",
        "grade/include/**/*.hpp",
        "template/**/*"]},
    include_package_data=True,
    zip_safe=False,
    install_requires=["jinja2", "jsonschema"])



import json
import jsonschema
from pathlib import Path
from typing import Any

from ..shared import *
from ..log import log
from .exception import CompilationException

root = Path(__file__).absolute().parent


class ValidationException(Exception):
    """Raised while building models."""


def resolve_local(uri: str) -> dict:
    """Resolve a locally referenced schema."""

    with root.joinpath("schema", uri).open() as file:
        return json.load(file)


def create_validator(schema_path: Path) -> jsonschema.Draft7Validator:
    """Create a validator with a local resolver.

    For convenience, we'll simply override the handler for refs
    passed without a protocol prefix. This allows PyCharm to
    understand the ref location and do path checking in the source
    schema file.
    """

    with schema_path.open() as file:
        schema = json.load(file)

    resolver = jsonschema.RefResolver(base_uri=str(schema_path), referrer=schema, handlers={"": resolve_local})
    return jsonschema.Draft7Validator(schema=schema, resolver=resolver)


assignment_validator = create_validator(root / "schema" / "assignment.schema.json")
problem_validator = create_validator(root / "schema" / "problem.schema.json")


def json_load_file(path: Path) -> Any:
    """Open a file and raise appropriate exceptions at on failure."""

    try:
        with path.open() as file:
            return json.load(file)
    except FileNotFoundError:
        raise CompilationException(message=f"No such path {path}")
    except (PermissionError, OSError):
        raise CompilationException(message=f"Failed to read {path}")
    except json.decoder.JSONDecodeError:
        raise CompilationException(message=f"Failed to deserialize {path}")


def validate_problem(problem_path: Path):
    """Validate with jsonschema."""

    data = json_load_file(problem_path.joinpath(Files.PROBLEM))

    try:
        problem_validator.validate(data)
    except jsonschema.ValidationError as exception:
        log.error(f"invalid schema in problem {problem_path}: {exception}")
        raise CompilationException(message=str(exception))


def validate_assignment(assignment_path: Path):
    """Validators per assignment."""

    data = json_load_file(assignment_path.joinpath(Files.ASSIGNMENT))

    try:
        assignment_validator.validate(data)
    except jsonschema.ValidationError as exception:
        log.error(f"invalid schema in assignment {assignment_path}: {exception}")
        raise CompilationException(message=str(exception))

    for problem_reference in data["problems"]:
        validate_problem(assignment_path.joinpath(problem_reference["path"]))


def validate(assignment_path: Path):
    """Do a bunch of checks """

    validate_assignment(assignment_path)



import abc

import jinja2

from pathlib import Path
from dataclasses import dataclass, field
from typing import Dict, Set, Optional, Type, List

from .models import CompilationAssignment
from ..log import log


@dataclass(repr=False, eq=False)
class Configuration:
    """Build context.

    The build context is simply a container for information about the
    build that's passed to each step.
    """

    assignment_path: Path
    artifacts_path: Path
    custom_template_path: Path
    options: Dict[str, str]

    def __post_init__(self):
        if self.custom_template_path is not None:
            log.info(f"custom template path is {self.custom_template_path}")


@dataclass(repr=False, eq=False)
class Context:
    """Details generated at compile time."""

    environment: jinja2.Environment
    indices_modified: bool = False
    paths_modified: Optional[Set[Path]] = None


@dataclass(repr=False, eq=False)
class UnitResult:
    """Compilation result."""

    compiled: bool = False
    tags: Set[str] = field(default_factory=set)


@dataclass(repr=False, eq=False)
class TargetResult:
    """Aggregated compilation results."""

    data: Dict[str, UnitResult] = field(default_factory=dict, init=False)

    def __setitem__(self, key: str, value: UnitResult):
        self.data[key] = value

    def __getitem__(self, item: str):
        return self.data[item]


class Unit(abc.ABC):
    """An independent component of the compilation process."""

    name: str
    configuration: Configuration

    def __init__(self, configuration: Configuration):
        self.configuration = configuration

    @abc.abstractmethod
    def compile(self, assignment: CompilationAssignment, context: Context) -> UnitResult:
        """Compile the unit given the context."""


class Workflow(abc.ABC):
    """Invoked after compilation."""

    configuration: Configuration

    def __init__(self, configuration: Configuration):
        self.configuration = configuration

    @abc.abstractmethod
    def run(self, assignment: CompilationAssignment, result: TargetResult):
        """Invoked after compilation."""


class Target:
    """A complete compilation target."""

    configuration: Configuration
    units: Dict[str, Unit]
    workflows: List[Workflow]

    def __init__(self, configuration: Configuration):
        """Set configuration reference."""

        self.configuration = configuration
        self.units = dict()
        self.workflows = list()

    def unit(self, unit_type: Type[Unit]):
        """Add a unit by its constructor."""

        self.units[unit_type.name] = unit_type(self.configuration)

    def workflow(self, workflow_type: Type[Workflow]):
        """Add a workflow."""

        self.workflows.append(workflow_type(self.configuration))

    def compile(self, assignment: CompilationAssignment, context: Context) -> TargetResult:
        """Compile and return results."""

        result = TargetResult()
        for unit_name, unit in self.units.items():
            result[unit_name] = unit.compile(assignment, context)
        for workflow in self.workflows:
            workflow.run(assignment, result)
        return result



import json
from pathlib import Path

from ...library.template import jinja2_create_environment

ROOT = Path(__file__).absolute().parent


RETURN = """<span class="punctuation">&crarr;</span><br>"""
SPACE = """<span class="punctuation">&middot;</span>"""


def show_whitespace(contents: str) -> str:
    return contents.replace(" ", SPACE).replace("\n", RETURN)


def out_from_runtime_stdout(report: dict) -> str:
    return report["runtime"]["stdout"]


def compare_output(template_path: Path, report_path: Path) -> str:
    """Generate a comparison of the expected output."""

    with report_path.open() as file:
        reports = json.load(file)

    environment = jinja2_create_environment(custom_template_path=template_path)
    difference_template = environment.get_template("template:grade/compare/difference.html")
    compare_template = environment.get_template("template:grade/compare/compare.html")

    content = f"<h1>{report_path.parts[-1]}</h1>\n<p>Correct on left, student on right</p>\n"
    for problem_short, report in reports.items():
        content += f"<h2><code>{problem_short}</code></h2>"
        for task_name, result in report.items():
            if result["complete"] and not result["passed"] and "error" in result:
                expected = result["error"]["expected"]
                if isinstance(expected, str):
                    expected = [expected]
                if expected is None:
                    expected = []
                received = result["error"]["received"]

                expected = list(map(show_whitespace, expected))
                received = show_whitespace(received) if received is not None else ""
                content += difference_template.render(task_name=task_name, expected=expected, received=received)

    return compare_template.render(content=content)



import json
import jsonschema
import re
from pathlib import Path
from typing import Callable

from ...shared import Files
from ..validate import assignment_validator, problem_validator


Validator = Callable[[str], bool]


def validate_not_empty(string: str) -> bool:
    if len(string) == 0:
        print("Must not be empty!")
        return False
    return True


def validate_datetime(string: str) -> bool:
    if re.match(r"^\d{4}-\d{1,2}-\d{1,2} \d{2}:\d{2}:\d{2}$", string) is None:
        print("Must be of the form YYYY-MM-DD HH:MM:SS!")
        return False
    return True


def validate_datetime_nullable(string: str) -> bool:
    if len(string) == 0:
        return True
    return validate_datetime(string)


def validate_integral(string: str) -> bool:
    if not string.isnumeric():
        print("Must be a positive integer!")
        return False
    return True


def validate_float(string: str) -> bool:
    try:
        float(string)
    except ValueError:
        return False
    return True


def validate_email(string: str) -> bool:
    if re.match(r"^.+@.+\..+$", string) is None:
        print("Invalid email!")
        return False
    return True


def validate_boolean(string: str) -> bool:
    if string.lower() not in ("y", "n", "yes", "no"):
        print("Must be (y)es or (n)o!")
        return False
    return True


def validate_weight(string: str) -> bool:
    if re.match(r"\d+(\.\d+)?", string) is None:
        print("Must be a float!")
        return False
    if not 0 <= float(string) <= 1:
        print("Must be between 0 and 1 inclusive!")
        return False
    return True


def validated_input(prompt: str = "", *validators: Validator) -> str:
    """Input until validated."""

    while True:
        response = input(prompt)
        if all(validator(response) for validator in validators):
            return response


def input_assignment_json() -> dict:
    """Encapsulate input logic."""

    while True:
        assignment_json = {
            "title": validated_input("Assignment title: ", validate_not_empty),
            "authors": [
                {
                    "name": validated_input("Author name: ", validate_not_empty),
                    "email": validated_input("Author email (address@domain.com): ", validate_email)
                }
            ],
            "dates": {
                "assigned": validated_input("Assigned (YYYY-MM-DD HH:MM:SS or empty): ", validate_datetime_nullable),
                "due": validated_input("Due (YYYY-MM-DD HH:MM:SS or empty): ", validate_datetime_nullable),
                "deadline": validated_input("Deadline (YYYY-MM-DD HH:MM:SS or empty): ", validate_datetime_nullable),
            },
            "problems": [],
            "grading": {
                "points": validated_input("Total points: ", validate_integral)
            }
        }

        for field in "assigned", "due", "deadline":
            if len(assignment_json["dates"][field]) == 0:
                assignment_json["dates"][field] = None

        try:
            assignment_validator.validate(assignment_json)
        except jsonschema.ValidationError as exception:
            print(exception)
        else:
            return assignment_json


def generate_assignment_interactive(assignment_path: Path):
    """Generate a new assignment."""

    try:
        assignment_json = input_assignment_json()
    except KeyboardInterrupt:
        print("Cancelling...")
        return

    assignment_path.mkdir(parents=True, exist_ok=True)
    with assignment_path.joinpath(Files.ASSIGNMENT).open("w") as file:
        json.dump(assignment_json, file, indent=2)

    print(f"Created assignment {assignment_path.parts[-1]}!")


def input_problem_json() -> dict:
    while True:
        problem_json = {
            "title": validated_input("Problem title: ", validate_not_empty),
            "authors": [
                {
                    "name": validated_input("Author name: ", validate_not_empty),
                    "email": validated_input("Author email (address@domain.com): ", validate_email)
                }
            ],
            "topics": [
                *map(str.strip, input("Optional topics (separated by comma): ").split(","))
            ],
            "grading": {
                "points": validated_input("Total points: ", validate_integral),
                "automated": None,
                "review": None,
                "manual": None,
            }
        }

        for category in "automated", "review", "manual":
            if validated_input(f"Requires {category} grading (y/n): ", validate_boolean).lower().startswith("y"):
                problem_json["grading"][category] = {
                    "weight": validated_input(f"Weight for {category}: ", validate_float)
                }

        try:
            problem_validator.validate(problem_json)
        except jsonschema.ValidationError as exception:
            print(exception)
        else:
            return problem_json


def generate_problem_interactive(assignment_path: Path, problem_relative_path: Path):
    """Generate an assignment within the assignment."""

    with assignment_path.joinpath(Files.ASSIGNMENT).open("r") as file:
        assignment_json = json.load(file)

    for existing_problem_json in assignment_json["problems"]:
        if Path(existing_problem_json["path"]).parts[-1] == problem_relative_path.parts[-1]:
            print("A problem with the same short is already in the assignment!")
            return

    try:
        problem_json = input_problem_json()
    except KeyboardInterrupt:
        print("Cancelling...")
        return

    while True:
        assignment_json["problems"].append({
            "path": str(problem_relative_path),
            "grading": {
                "weight": validated_input("Problem weight: ", validate_float)
            }
        })

        try:
            assignment_validator.validate(assignment_json)
        except jsonschema.ValidationError as exception:
            print(exception)
        else:
            break

    with assignment_path.joinpath(Files.ASSIGNMENT).open("w") as file:
        json.dump(assignment_json, file, indent=2)

    problem_path = assignment_path.joinpath(problem_relative_path)
    problem_path.mkdir(parents=True, exist_ok=True)
    with problem_path.joinpath(Files.PROBLEM).open("w") as file:
        json.dump(problem_json, file, indent=2)

    print(f"Created problem {problem_relative_path}!")



import json
from pathlib import Path

from curricula.library.printer import Printer
from ..models import GradingAssignment
from ..report import AssignmentReport


def get_diagnostics(assignment: GradingAssignment, assignment_report_path: Path) -> str:
    """Check if tests passed, displaying errors."""

    # Load the assignment
    with assignment_report_path.open() as file:
        assignment_report = AssignmentReport.load(json.load(file), assignment)

    # Create buffer
    output = Printer()

    # Iterate problems
    for problem in assignment.problems:
        if problem.short not in assignment_report:
            continue

        problem_report = assignment_report[problem.short]
        results_passing_count = sum(1 for _ in filter(lambda p: p.passing, problem_report.results.values()))

        # Problem summary header
        output.print(f"Problem {problem.short}: {results_passing_count}/{len(problem_report.results.values())}")

        # Iterate results
        output.indent()
        for task in problem.grader.tasks:
            result = problem_report.get(task.name)
            if result is None:
                continue

            if not result.complete:
                output.print(f"✗ {task.name} did not complete")
            elif not result.passing:
                output.print(f"✗ {task.name} failed")
                output.indent()
                if result.error:
                    if result.error.location:
                        output.print(f"Reason: {result.error.description}")
                        output.print(f"Location: {result.error.location}")
                    if result.error.traceback:
                        output.print("Traceback: ")
                        output.print(result.error.traceback.strip(), indentation=2)
                    if result.error.expected:
                        output.print(f"Expected: {repr(result.error.expected)}")
                    if result.error.received:
                        output.print(f"Received: {repr(result.error.received)}")
                output.dedent()
            else:
                output.print(f"✓ {task.name} passed")

        output.dedent()
    return str(output)



import json

from ..models import CompilationAssignment
from ..framework import Workflow, TargetResult
from ...shared import Paths


class GradeWorkflow(Workflow):
    """Grades the solution and writes the report to the artifact root."""

    def run(self, assignment: CompilationAssignment, result: TargetResult):
        """Main."""

        if not self.configuration.options.get("grade"):
            return

        if not result["grading"].compiled and not result["solution"].compiled:
            return

        from ...grade import run
        from ...grade.models import GradingAssignment
        from ...grade.tools.format import format_report_markdown
        from ...grade.tools.diagnostics import get_diagnostics

        grading_path = self.configuration.artifacts_path.joinpath(Paths.GRADING)
        solution_path = self.configuration.artifacts_path.joinpath(Paths.SOLUTION)
        report_path = self.configuration.artifacts_path.joinpath("solution.report.json")
        formatted_report_path = self.configuration.artifacts_path.joinpath("solution.report.md")

        assignment = GradingAssignment.read(grading_path)
        report = run(assignment, solution_path, options=self.configuration.options)

        with report_path.open("w") as file:
            json.dump(report.dump(), file, indent=2)

        with formatted_report_path.open("w") as file:
            file.write(format_report_markdown(
                assignment=assignment,
                report_path=report_path,
                options=self.configuration.options))

        get_diagnostics(assignment, report_path)



from dataclasses import dataclass

from ...task import Result, Error


@dataclass
class CleanupResult(Result):
    """Deletion of files, deallocation, etc."""

    kind = "cleanup"

    def __init__(self, passing: bool = True, complete: bool = True, error: Error = None, **details):
        super().__init__(complete=complete, passing=passing, error=error, details=details)
